@@ -6,15 +6,19 @@
 //! such as `rpm-ostree` can directly reuse it.
 
 use anyhow::Result;
-use ostree::{gio, glib};
+use futures_util::FutureExt;
+use ostree::{cap_std, gio, glib};
+use std::borrow::Borrow;
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::ffi::OsString;
 use std::path::PathBuf;
 use structopt::StructOpt;
+use tokio_stream::StreamExt;
 
-use crate::container as ostree_container;
+use crate::commit::container_commit;
 use crate::container::store::{LayeredImageImporter, PrepareResult};
+use crate::container::{self as ostree_container, UnencapsulationProgress};
 use crate::container::{Config, ImageReference, OstreeImageReference, UnencapsulateOptions};
 
 fn parse_imgref(s: &str) -> Result<OstreeImageReference> {
@@ -25,12 +29,18 @@ fn parse_base_imgref(s: &str) -> Result<ImageReference> {
     ImageReference::try_from(s)
 }
 
+fn parse_repo(s: &str) -> Result<ostree::Repo> {
+    let repofd = cap_std::fs::Dir::open_ambient_dir(s, cap_std::ambient_authority())?;
+    Ok(ostree::Repo::open_at_dir(&repofd, ".")?)
+}
+
 /// Options for importing a tar archive.
 #[derive(Debug, StructOpt)]
 struct ImportOpts {
     /// Path to the repository
     #[structopt(long)]
-    repo: String,
+    #[structopt(parse(try_from_str = parse_repo))]
+    repo: ostree::Repo,
 
     /// Path to a tar archive; if unspecified, will be stdin.  Currently the tar archive must not be compressed.
     path: Option<String>,
@@ -41,7 +51,8 @@ struct ImportOpts {
 struct ExportOpts {
     /// Path to the repository
     #[structopt(long)]
-    repo: String,
+    #[structopt(parse(try_from_str = parse_repo))]
+    repo: ostree::Repo,
 
     /// The format version.  Must be 0 or 1.
     #[structopt(long)]
@@ -69,7 +80,8 @@ enum ContainerOpts {
     Unencapsulate {
         /// Path to the repository
         #[structopt(long)]
-        repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        repo: ostree::Repo,
 
         /// Image reference, e.g. registry:quay.io/exampleos/exampleos:latest
         #[structopt(parse(try_from_str = parse_imgref))]
@@ -96,7 +108,8 @@ enum ContainerOpts {
     Encapsulate {
         /// Path to the repository
         #[structopt(long)]
-        repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        repo: ostree::Repo,
 
         /// The ostree ref or commit to export
         rev: String,
@@ -109,11 +122,18 @@ enum ContainerOpts {
         #[structopt(name = "label", long, short)]
         labels: Vec<String>,
 
+        /// Propagate an OSTree commit metadata key to container label
+        #[structopt(name = "copymeta", long)]
+        copy_meta_keys: Vec<String>,
+
         /// Corresponds to the Dockerfile `CMD` instruction.
         #[structopt(long)]
         cmd: Option<Vec<String>>,
     },
 
+    #[structopt(alias = "commit")]
+    Commit,
+
     /// Commands for working with (possibly layered, non-encapsulated) container images.
     Image(ContainerImageOpts),
 }
@@ -121,6 +141,10 @@ enum ContainerOpts {
 /// Options for container image fetching.
 #[derive(Debug, StructOpt)]
 struct ContainerProxyOpts {
+    #[structopt(long)]
+    /// Do not use default authentication files.
+    auth_anonymous: bool,
+
     #[structopt(long)]
     /// Path to Docker-formatted authentication file.
     authfile: Option<PathBuf>,
@@ -142,14 +166,15 @@ enum ContainerImageOpts {
     List {
         /// Path to the repository
         #[structopt(long)]
-        repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        repo: ostree::Repo,
     },
 
     /// Pull (or update) a container image.
     Pull {
         /// Path to the repository
-        #[structopt(long)]
-        repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        repo: ostree::Repo,
 
         /// Image reference, e.g. ostree-remote-image:someremote:registry:quay.io/exampleos/exampleos:latest
         #[structopt(parse(try_from_str = parse_imgref))]
@@ -163,11 +188,13 @@ enum ContainerImageOpts {
     Copy {
         /// Path to the source repository
         #[structopt(long)]
-        src_repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        src_repo: ostree::Repo,
 
         /// Path to the destination repository
         #[structopt(long)]
-        dest_repo: String,
+        #[structopt(parse(try_from_str = parse_repo))]
+        dest_repo: ostree::Repo,
 
         /// Image reference, e.g. ostree-remote-image:someremote:registry:quay.io/exampleos/exampleos:latest
         #[structopt(parse(try_from_str = parse_imgref))]
@@ -211,7 +238,8 @@ enum ContainerImageOpts {
 struct ImaSignOpts {
     /// Path to the repository
     #[structopt(long)]
-    repo: String,
+    #[structopt(parse(try_from_str = parse_repo))]
+    repo: ostree::Repo,
     /// The ostree ref or commit to use as a base
     src_rev: String,
     /// The ostree ref to use for writing the signed commit
@@ -228,6 +256,9 @@ struct ImaSignOpts {
 enum TestingOpts {
     // Detect the current environment
     DetectEnv,
+    /// Execute integration tests, assuming mutable environment
+    Run,
+    FilterTar,
 }
 
 /// Toplevel options for extended ostree functionality.
@@ -251,6 +282,7 @@ enum Opt {
 impl Into<ostree_container::store::ImageProxyConfig> for ContainerProxyOpts {
     fn into(self) -> ostree_container::store::ImageProxyConfig {
         ostree_container::store::ImageProxyConfig {
+            auth_anonymous: self.auth_anonymous,
             authfile: self.authfile,
             certificate_directory: self.cert_dir,
             insecure_skip_tls_verification: Some(self.insecure_skip_tls_verification),
@@ -261,13 +293,12 @@ impl Into<ostree_container::store::ImageProxyConfig> for ContainerProxyOpts {
 
 /// Import a tar archive containing an ostree commit.
 async fn tar_import(opts: &ImportOpts) -> Result<()> {
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, opts.repo.as_str(), gio::NONE_CANCELLABLE)?;
     let imported = if let Some(path) = opts.path.as_ref() {
         let instream = tokio::fs::File::open(path).await?;
-        crate::tar::import_tar(repo, instream, None).await?
+        crate::tar::import_tar(&opts.repo, instream, None).await?
     } else {
         let stdin = tokio::io::stdin();
-        crate::tar::import_tar(repo, stdin, None).await?
+        crate::tar::import_tar(&opts.repo, stdin, None).await?
     };
     println!("Imported: {}", imported);
     Ok(())
@@ -275,24 +306,32 @@ async fn tar_import(opts: &ImportOpts) -> Result<()> {
 
 /// Export a tar archive containing an ostree commit.
 fn tar_export(opts: &ExportOpts) -> Result<()> {
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, opts.repo.as_str(), gio::NONE_CANCELLABLE)?;
     #[allow(clippy::needless_update)]
     let subopts = crate::tar::ExportOptions {
         format_version: opts.format_version,
         ..Default::default()
     };
-    crate::tar::export_commit(repo, opts.rev.as_str(), std::io::stdout(), Some(subopts))?;
+    crate::tar::export_commit(
+        &opts.repo,
+        opts.rev.as_str(),
+        std::io::stdout(),
+        Some(subopts),
+    )?;
     Ok(())
 }
 
+enum ProgressOrFinish {
+    Progress(UnencapsulationProgress),
+    Finished(Result<ostree_container::Import>),
+}
+
 /// Import a container image with an encapsulated ostree commit.
 async fn container_import(
-    repo: &str,
+    repo: &ostree::Repo,
     imgref: &OstreeImageReference,
     write_ref: Option<&str>,
     quiet: bool,
 ) -> Result<()> {
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, repo, gio::NONE_CANCELLABLE)?;
     let (tx_progress, rx_progress) = tokio::sync::watch::channel(Default::default());
     let target = indicatif::ProgressDrawTarget::stdout();
     let style = indicatif::ProgressStyle::default_bar();
@@ -309,26 +348,32 @@ async fn container_import(
     let opts = UnencapsulateOptions {
         progress: Some(tx_progress),
     };
-    let import = crate::container::unencapsulate(repo, imgref, Some(opts));
-    tokio::pin!(import);
-    tokio::pin!(rx_progress);
-    let import = loop {
-        tokio::select! {
-            _ = rx_progress.changed() => {
-                let n = rx_progress.borrow().processed_bytes;
+    let rx_progress_stream =
+        tokio_stream::wrappers::WatchStream::new(rx_progress).map(ProgressOrFinish::Progress);
+    let import = crate::container::unencapsulate(repo, imgref, Some(opts))
+        .into_stream()
+        .map(ProgressOrFinish::Finished);
+    let stream = rx_progress_stream.merge(import);
+    tokio::pin!(stream);
+    let mut import_result = None;
+    while let Some(value) = stream.next().await {
+        match value {
+            ProgressOrFinish::Progress(progress) => {
+                let n = progress.borrow().processed_bytes;
                 if let Some(pb) = pb.as_ref() {
                     pb.set_message(format!("Processed: {}", indicatif::HumanBytes(n)));
                 }
             }
-            import = &mut import => {
-                if let Some(pb) = pb.as_ref() {
-                    pb.finish();
-                }
-                break import?;
+            ProgressOrFinish::Finished(import) => {
+                import_result = Some(import?);
             }
         }
-    };
-
+    }
+    if let Some(pb) = pb.as_ref() {
+        pb.finish();
+    }
+    // It must have been set
+    let import = import_result.unwrap();
     if let Some(write_ref) = write_ref {
         repo.set_ref_immediate(
             None,
@@ -350,19 +395,22 @@ async fn container_import(
 
 /// Export a container image with an encapsulated ostree commit.
 async fn container_export(
-    repo: &str,
+    repo: &ostree::Repo,
     rev: &str,
     imgref: &ImageReference,
     labels: BTreeMap<String, String>,
+    copy_meta_keys: Vec<String>,
     cmd: Option<Vec<String>>,
 ) -> Result<()> {
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, repo, gio::NONE_CANCELLABLE)?;
     let config = Config {
         labels: Some(labels),
         cmd,
     };
-    let opts = Some(Default::default());
-    let pushed = crate::container::encapsulate(repo, rev, &config, opts, imgref).await?;
+    let opts = crate::container::ExportOpts {
+        copy_meta_keys,
+        ..Default::default()
+    };
+    let pushed = crate::container::encapsulate(repo, rev, &config, Some(opts), imgref).await?;
     println!("{}", pushed);
     Ok(())
 }
@@ -376,11 +424,10 @@ async fn container_info(imgref: &OstreeImageReference) -> Result<()> {
 
 /// Write a layered container image into an OSTree commit.
 async fn container_store(
-    repo: &str,
+    repo: &ostree::Repo,
     imgref: &OstreeImageReference,
     proxyopts: ContainerProxyOpts,
 ) -> Result<()> {
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, repo, gio::NONE_CANCELLABLE)?;
     let mut imp = LayeredImageImporter::new(repo, imgref, proxyopts.into()).await?;
     let prep = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(c) => {
@@ -427,14 +474,12 @@ async fn container_store(
 
 /// Add IMA signatures to an ostree commit, generating a new commit.
 fn ima_sign(cmdopts: &ImaSignOpts) -> Result<()> {
-    let repo =
-        &ostree::Repo::open_at(libc::AT_FDCWD, cmdopts.repo.as_str(), gio::NONE_CANCELLABLE)?;
     let signopts = crate::ima::ImaOpts {
         algorithm: cmdopts.algorithm.clone(),
         key: cmdopts.key.clone(),
     };
-    let signed_commit = crate::ima::ima_sign(repo, cmdopts.src_rev.as_str(), &signopts)?;
-    repo.set_ref_immediate(
+    let signed_commit = crate::ima::ima_sign(&cmdopts.repo, cmdopts.src_rev.as_str(), &signopts)?;
+    cmdopts.repo.set_ref_immediate(
         None,
         cmdopts.target_ref.as_str(),
         Some(signed_commit.as_str()),
@@ -452,6 +497,10 @@ fn testing(opts: &TestingOpts) -> Result<()> {
             println!("{}", s);
             Ok(())
         }
+        TestingOpts::Run => crate::integrationtest::run_tests(),
+        TestingOpts::FilterTar => {
+            crate::tar::filter_tar(std::io::stdin(), std::io::stdout()).map(|_| {})
+        }
     }
 }
 
@@ -468,6 +517,7 @@ where
         Opt::Tar(TarOpts::Export(ref opt)) => tar_export(opt),
         Opt::Container(o) => match o {
             ContainerOpts::Info { imgref } => container_info(&imgref).await,
+            ContainerOpts::Commit {} => container_commit().await,
             ContainerOpts::Unencapsulate {
                 repo,
                 imgref,
@@ -479,6 +529,7 @@ where
                 rev,
                 imgref,
                 labels,
+                copy_meta_keys,
                 cmd,
             } => {
                 let labels: Result<BTreeMap<_, _>> = labels
@@ -490,13 +541,11 @@ where
                         Ok((k.to_string(), v.to_string()))
                     })
                     .collect();
-                container_export(&repo, &rev, &imgref, labels?, cmd).await
+                container_export(&repo, &rev, &imgref, labels?, copy_meta_keys, cmd).await
             }
             ContainerOpts::Image(opts) => match opts {
                 ContainerImageOpts::List { repo } => {
-                    let repo =
-                        &ostree::Repo::open_at(libc::AT_FDCWD, &repo, gio::NONE_CANCELLABLE)?;
-                    for image in crate::container::store::list_images(repo)? {
+                    for image in crate::container::store::list_images(&repo)? {
                         println!("{}", image);
                     }
                     Ok(())
@@ -510,13 +559,7 @@ where
                     src_repo,
                     dest_repo,
                     imgref,
-                } => {
-                    let src_repo =
-                        &ostree::Repo::open_at(libc::AT_FDCWD, &src_repo, gio::NONE_CANCELLABLE)?;
-                    let dest_repo =
-                        &ostree::Repo::open_at(libc::AT_FDCWD, &dest_repo, gio::NONE_CANCELLABLE)?;
-                    crate::container::store::copy(src_repo, dest_repo, &imgref).await
-                }
+                } => crate::container::store::copy(&src_repo, &dest_repo, &imgref).await,
                 ContainerImageOpts::Deploy {
                     sysroot,
                     stateroot,
