@@ -1,7 +1,7 @@
+mod fixture;
+
 use anyhow::{Context, Result};
-use camino::{Utf8Path, Utf8PathBuf};
-use fn_error_context::context;
-use indoc::indoc;
+use camino::Utf8Path;
 use once_cell::sync::Lazy;
 use ostree_ext::container::store::PrepareResult;
 use ostree_ext::container::{
@@ -9,18 +9,13 @@ use ostree_ext::container::{
 };
 use ostree_ext::tar::TarImportOptions;
 use ostree_ext::{gio, glib};
-use sh_inline::bash;
+use sh_inline::bash_in;
 use std::collections::HashMap;
-use std::convert::TryInto;
-use std::{io::Write, process::Command};
+use std::process::Command;
 
-const OSTREE_GPG_HOME: &[u8] = include_bytes!("fixtures/ostree-gpg-test-home.tar.gz");
-const TEST_GPG_KEYID_1: &str = "7FCA23D8472CDAFA";
-#[allow(dead_code)]
-const TEST_GPG_KEYFPR_1: &str = "5E65DE75AB1C501862D476347FCA23D8472CDAFA";
-const EXAMPLEOS_V0: &[u8] = include_bytes!("fixtures/exampleos.tar.zst");
-const EXAMPLEOS_V1: &[u8] = include_bytes!("fixtures/exampleos-v1.tar.zst");
-const TESTREF: &str = "exampleos/x86_64/stable";
+use fixture::Fixture;
+
+const EXAMPLE_TAR_LAYER: &[u8] = include_bytes!("fixtures/hlinks.tar.gz");
 const EXAMPLEOS_CONTENT_CHECKSUM: &str =
     "0ef7461f9db15e1d8bd8921abf20694225fbaa4462cadf7deed8ea0e43162120";
 const TEST_REGISTRY_DEFAULT: &str = "localhost:5000";
@@ -38,139 +33,30 @@ static TEST_REGISTRY: Lazy<String> = Lazy::new(|| match std::env::var_os("TEST_R
     None => TEST_REGISTRY_DEFAULT.to_string(),
 });
 
-#[context("Generating test repo")]
-fn generate_test_repo(dir: &Utf8Path) -> Result<Utf8PathBuf> {
-    let src_tarpath = &dir.join("exampleos.tar.zst");
-    std::fs::write(src_tarpath, EXAMPLEOS_V0)?;
-
-    let gpghome = dir.join("gpghome");
-    {
-        let dec = flate2::read::GzDecoder::new(OSTREE_GPG_HOME);
-        let mut a = tar::Archive::new(dec);
-        a.unpack(&gpghome)?;
-    };
-
-    bash!(
-        indoc! {"
-        cd {dir}
-        ostree --repo=repo init --mode=archive
-        ostree --repo=repo commit -b {testref} --bootable --no-bindings --add-metadata-string=version=42.0 --gpg-homedir={gpghome} --gpg-sign={keyid} \
-               --add-detached-metadata-string=my-detached-key=my-detached-value --tree=tar=exampleos.tar.zst >/dev/null
-        ostree --repo=repo show {testref} >/dev/null
-    "},
-        testref = TESTREF,
-        gpghome = gpghome.as_str(),
-        keyid = TEST_GPG_KEYID_1,
-        dir = dir.as_str()
-    )?;
-    std::fs::remove_file(src_tarpath)?;
-    Ok(dir.join("repo"))
-}
-
-fn update_repo(repopath: &Utf8Path) -> Result<()> {
-    let repotmp = &repopath.join("tmp");
-    let srcpath = &repotmp.join("exampleos-v1.tar.zst");
-    std::fs::write(srcpath, EXAMPLEOS_V1)?;
-    let srcpath = srcpath.as_str();
-    let repopath = repopath.as_str();
-    let testref = TESTREF;
-    bash!(
-        "ostree --repo={repopath} commit -b {testref} --no-bindings --tree=tar={srcpath}",
-        testref,
-        repopath,
-        srcpath
-    )?;
-    std::fs::remove_file(srcpath)?;
-    Ok(())
-}
-
-#[context("Generating test tarball")]
-fn initial_export(fixture: &Fixture) -> Result<Utf8PathBuf> {
-    let cancellable = gio::NONE_CANCELLABLE;
-    let (_, rev) = fixture.srcrepo.read_commit(TESTREF, cancellable)?;
-    let (commitv, _) = fixture.srcrepo.load_commit(rev.as_str())?;
-    assert_eq!(
-        ostree::commit_get_content_checksum(&commitv)
-            .unwrap()
-            .as_str(),
-        EXAMPLEOS_CONTENT_CHECKSUM
-    );
-    let destpath = fixture.path.join("exampleos-export.tar");
-    let mut outf = std::io::BufWriter::new(std::fs::File::create(&destpath)?);
-    let options = ostree_ext::tar::ExportOptions {
-        format_version: fixture.format_version,
-        ..Default::default()
-    };
-    ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut outf, Some(options))?;
-    outf.flush()?;
-    Ok(destpath)
-}
-
-struct Fixture {
-    // Just holds a reference
-    _tempdir: tempfile::TempDir,
-    path: Utf8PathBuf,
-    srcdir: Utf8PathBuf,
-    srcrepo: ostree::Repo,
-    destrepo: ostree::Repo,
-    destrepo_path: Utf8PathBuf,
-
-    format_version: u32,
-}
-
-impl Fixture {
-    fn new() -> Result<Self> {
-        let _tempdir = tempfile::tempdir_in("/var/tmp")?;
-        let path: &Utf8Path = _tempdir.path().try_into().unwrap();
-        let path = path.to_path_buf();
-
-        let srcdir = path.join("src");
-        std::fs::create_dir(&srcdir)?;
-        let srcrepo_path = generate_test_repo(&srcdir)?;
-        let srcrepo =
-            ostree::Repo::open_at(libc::AT_FDCWD, srcrepo_path.as_str(), gio::NONE_CANCELLABLE)?;
-
-        let destdir = &path.join("dest");
-        std::fs::create_dir(destdir)?;
-        let destrepo_path = destdir.join("repo");
-        let destrepo = ostree::Repo::new_for_path(&destrepo_path);
-        destrepo.create(ostree::RepoMode::BareUser, gio::NONE_CANCELLABLE)?;
-        Ok(Self {
-            _tempdir,
-            path,
-            srcdir,
-            srcrepo,
-            destrepo,
-            destrepo_path,
-            format_version: 0,
-        })
-    }
-}
-
 #[tokio::test]
 async fn test_tar_import_empty() -> Result<()> {
     let fixture = Fixture::new()?;
-    let r = ostree_ext::tar::import_tar(&fixture.destrepo, tokio::io::empty(), None).await;
+    let r = ostree_ext::tar::import_tar(fixture.destrepo(), tokio::io::empty(), None).await;
     assert_err_contains(r, "Commit object not found");
     Ok(())
 }
 
 #[tokio::test]
 async fn test_tar_export_reproducible() -> Result<()> {
-    let fixture = Fixture::new()?;
+    let fixture = Fixture::new_v1()?;
     let (_, rev) = fixture
-        .srcrepo
-        .read_commit(TESTREF, gio::NONE_CANCELLABLE)?;
+        .srcrepo()
+        .read_commit(fixture.testref(), gio::NONE_CANCELLABLE)?;
     let export1 = {
         let mut h = openssl::hash::Hasher::new(openssl::hash::MessageDigest::sha256())?;
-        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h, None)?;
+        ostree_ext::tar::export_commit(fixture.srcrepo(), rev.as_str(), &mut h, None)?;
         h.finish()?
     };
     // Artificial delay to flush out mtimes (one second granularity baseline, plus another 100ms for good measure).
     std::thread::sleep(std::time::Duration::from_millis(1100));
     let export2 = {
         let mut h = openssl::hash::Hasher::new(openssl::hash::MessageDigest::sha256())?;
-        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h, None)?;
+        ostree_ext::tar::export_commit(fixture.srcrepo(), rev.as_str(), &mut h, None)?;
         h.finish()?
     };
     assert_eq!(*export1, *export2);
@@ -180,12 +66,21 @@ async fn test_tar_export_reproducible() -> Result<()> {
 #[tokio::test]
 async fn test_tar_import_signed() -> Result<()> {
     let fixture = Fixture::new()?;
-    let test_tar = &initial_export(&fixture)?;
+    let test_tar = fixture.export_tar()?;
+
+    let rev = fixture.srcrepo().require_rev(fixture.testref())?;
+    let (commitv, _) = fixture.srcrepo().load_commit(rev.as_str())?;
+    assert_eq!(
+        ostree::commit_get_content_checksum(&commitv)
+            .unwrap()
+            .as_str(),
+        EXAMPLEOS_CONTENT_CHECKSUM
+    );
 
     // Verify we fail with an unknown remote.
-    let src_tar = tokio::fs::File::open(test_tar).await?;
+    let src_tar = tokio::fs::File::from_std(fixture.dir.open(test_tar)?.into_std());
     let r = ostree_ext::tar::import_tar(
-        &fixture.destrepo,
+        fixture.destrepo(),
         src_tar,
         Some(TarImportOptions {
             remote: Some("nosuchremote".to_string()),
@@ -199,11 +94,11 @@ async fn test_tar_import_signed() -> Result<()> {
     opts.insert("gpg-verify", &true);
     opts.insert("custom-backend", &"ostree-rs-ext");
     fixture
-        .destrepo
+        .destrepo()
         .remote_add("myremote", None, Some(&opts.end()), gio::NONE_CANCELLABLE)?;
-    let src_tar = tokio::fs::File::open(test_tar).await?;
+    let src_tar = tokio::fs::File::from_std(fixture.dir.open(test_tar)?.into_std());
     let r = ostree_ext::tar::import_tar(
-        &fixture.destrepo,
+        fixture.destrepo(),
         src_tar,
         Some(TarImportOptions {
             remote: Some("myremote".to_string()),
@@ -213,21 +108,19 @@ async fn test_tar_import_signed() -> Result<()> {
     assert_err_contains(r, r#"Can't check signature: public key not found"#);
 
     // And signed correctly
-    bash!(
-        "ostree --repo={repo} remote gpg-import --stdin myremote < {p}/gpghome/key1.asc >/dev/null",
-        repo = fixture.destrepo_path.as_str(),
-        p = fixture.srcdir.as_str()
+    bash_in!(&fixture.dir,
+        "ostree --repo=dest/repo remote gpg-import --stdin myremote < src/gpghome/key1.asc >/dev/null",
     )?;
-    let src_tar = tokio::fs::File::open(test_tar).await?;
+    let src_tar = tokio::fs::File::from_std(fixture.dir.open(test_tar)?.into_std());
     let imported = ostree_ext::tar::import_tar(
-        &fixture.destrepo,
+        fixture.destrepo(),
         src_tar,
         Some(TarImportOptions {
             remote: Some("myremote".to_string()),
         }),
     )
     .await?;
-    let (commitdata, state) = fixture.destrepo.load_commit(&imported)?;
+    let (commitdata, state) = fixture.destrepo().load_commit(&imported)?;
     assert_eq!(
         EXAMPLEOS_CONTENT_CHECKSUM,
         ostree::commit_get_content_checksum(&commitdata)
@@ -256,6 +149,7 @@ impl Into<TarExpected> for &(&'static str, tar::EntryType, u32) {
 }
 
 fn validate_tar_expected<T: std::io::Read>(
+    format_version: u32,
     t: tar::Entries<T>,
     expected: impl IntoIterator<Item = TarExpected>,
 ) -> Result<()> {
@@ -269,7 +163,25 @@ fn validate_tar_expected<T: std::io::Read>(
         let entry_path = entry.path().unwrap().to_string_lossy().into_owned();
         if let Some(exp) = expected.remove(entry_path.as_str()) {
             assert_eq!(header.entry_type(), exp.etype, "{}", entry_path);
-            assert_eq!(header.mode().unwrap(), exp.mode, "{}", entry_path);
+            let is_old_object = format_version == 0;
+            let mut expected_mode = exp.mode;
+            if is_old_object && !entry_path.starts_with("sysroot/") {
+                let fmtbits = match header.entry_type() {
+                    tar::EntryType::Regular => libc::S_IFREG,
+                    tar::EntryType::Directory => libc::S_IFDIR,
+                    tar::EntryType::Symlink => 0,
+                    o => panic!("Unexpected entry type {:?}", o),
+                };
+                expected_mode |= fmtbits;
+            }
+            assert_eq!(
+                header.mode().unwrap(),
+                expected_mode,
+                "fmtver: {} type: {:?} path: {}",
+                format_version,
+                header.entry_type(),
+                entry_path
+            );
         }
     }
 
@@ -287,15 +199,15 @@ fn test_tar_export_structure() -> Result<()> {
     use tar::EntryType::{Directory, Regular};
 
     let mut fixture = Fixture::new()?;
-    let src_tar = initial_export(&fixture)?;
-    let src_tar = std::io::BufReader::new(std::fs::File::open(&src_tar)?);
+    let src_tar = fixture.export_tar()?;
+    let src_tar = std::io::BufReader::new(fixture.dir.open(src_tar)?);
     let mut src_tar = tar::Archive::new(src_tar);
     let mut entries = src_tar.entries()?;
     // The first entry should be the root directory.
     let first = entries.next().unwrap()?;
     let firstpath = first.path()?;
     assert_eq!(firstpath.to_str().unwrap(), "./");
-    assert_eq!(first.header().mode()?, 0o755);
+    assert_eq!(first.header().mode()?, libc::S_IFDIR | 0o755);
     let next = entries.next().unwrap().unwrap();
     assert_eq!(next.path().unwrap().as_os_str(), "sysroot");
 
@@ -303,6 +215,7 @@ fn test_tar_export_structure() -> Result<()> {
     let expected = [
         ("sysroot/config", Regular, 0o644),
         ("sysroot/ostree/repo", Directory, 0o755),
+        ("sysroot/ostree/repo/extensions", Directory, 0o755),
         ("sysroot/ostree/repo/objects/00", Directory, 0o755),
         ("sysroot/ostree/repo/objects/23", Directory, 0o755),
         ("sysroot/ostree/repo/objects/77", Directory, 0o755),
@@ -313,21 +226,27 @@ fn test_tar_export_structure() -> Result<()> {
         ("sysroot/ostree/repo/refs/heads", Directory, 0o755),
         ("sysroot/ostree/repo/refs/mirrors", Directory, 0o755),
         ("sysroot/ostree/repo/refs/remotes", Directory, 0o755),
+        ("sysroot/ostree/repo/state", Directory, 0o755),
         ("sysroot/ostree/repo/tmp", Directory, 0o755),
         ("sysroot/ostree/repo/tmp/cache", Directory, 0o755),
         ("sysroot/ostree/repo/xattrs", Directory, 0o755),
         ("usr", Directory, 0o755),
     ];
-    validate_tar_expected(entries, expected.iter().map(Into::into))?;
+    validate_tar_expected(
+        fixture.format_version,
+        entries,
+        expected.iter().map(Into::into),
+    )?;
 
     // Validate format version 1
     fixture.format_version = 1;
-    let src_tar = initial_export(&fixture)?;
-    let src_tar = std::io::BufReader::new(std::fs::File::open(&src_tar)?);
+    let src_tar = fixture.export_tar()?;
+    let src_tar = std::io::BufReader::new(fixture.dir.open(src_tar)?);
     let mut src_tar = tar::Archive::new(src_tar);
     let expected = [
         ("sysroot/ostree/repo", Directory, 0o755),
         ("sysroot/ostree/repo/config", Regular, 0o644),
+        ("sysroot/ostree/repo/extensions", Directory, 0o755),
         ("sysroot/ostree/repo/objects/00", Directory, 0o755),
         ("sysroot/ostree/repo/objects/23", Directory, 0o755),
         ("sysroot/ostree/repo/objects/77", Directory, 0o755),
@@ -338,12 +257,16 @@ fn test_tar_export_structure() -> Result<()> {
         ("sysroot/ostree/repo/refs/heads", Directory, 0o755),
         ("sysroot/ostree/repo/refs/mirrors", Directory, 0o755),
         ("sysroot/ostree/repo/refs/remotes", Directory, 0o755),
+        ("sysroot/ostree/repo/state", Directory, 0o755),
         ("sysroot/ostree/repo/tmp", Directory, 0o755),
         ("sysroot/ostree/repo/tmp/cache", Directory, 0o755),
-        ("sysroot/ostree/repo/xattrs", Directory, 0o755),
         ("usr", Directory, 0o755),
     ];
-    validate_tar_expected(src_tar.entries()?, expected.iter().map(Into::into))?;
+    validate_tar_expected(
+        fixture.format_version,
+        src_tar.entries()?,
+        expected.iter().map(Into::into),
+    )?;
 
     Ok(())
 }
@@ -351,24 +274,25 @@ fn test_tar_export_structure() -> Result<()> {
 #[tokio::test]
 async fn test_tar_import_export() -> Result<()> {
     let fixture = Fixture::new()?;
-    let src_tar = tokio::fs::File::open(&initial_export(&fixture)?).await?;
+    let p = fixture.export_tar()?;
+    let src_tar = tokio::fs::File::from_std(fixture.dir.open(p)?.into_std());
 
     let imported_commit: String =
-        ostree_ext::tar::import_tar(&fixture.destrepo, src_tar, None).await?;
-    let (commitdata, _) = fixture.destrepo.load_commit(&imported_commit)?;
+        ostree_ext::tar::import_tar(fixture.destrepo(), src_tar, None).await?;
+    let (commitdata, _) = fixture.destrepo().load_commit(&imported_commit)?;
     assert_eq!(
         EXAMPLEOS_CONTENT_CHECKSUM,
         ostree::commit_get_content_checksum(&commitdata)
             .unwrap()
             .as_str()
     );
-    bash!(
+    bash_in!(
+        &fixture.dir,
         r#"
-         ostree --repo={destrepodir} ls -R {imported_commit} >/dev/null
-         val=$(ostree --repo={destrepodir} show --print-detached-metadata-key=my-detached-key {imported_commit})
-         test "${{val}}" = "'my-detached-value'"
+         ostree --repo=dest/repo ls -R ${imported_commit} >/dev/null
+         val=$(ostree --repo=dest/repo show --print-detached-metadata-key=my-detached-key ${imported_commit})
+         test "${val}" = "'my-detached-value'"
         "#,
-        destrepodir = fixture.destrepo_path.as_str(),
         imported_commit = imported_commit.as_str()
     )?;
     Ok(())
@@ -378,26 +302,24 @@ async fn test_tar_import_export() -> Result<()> {
 async fn test_tar_write() -> Result<()> {
     let fixture = Fixture::new()?;
     // Test translating /etc to /usr/etc
-    let tmpetc = fixture.path.join("tmproot/etc");
-    std::fs::create_dir_all(&tmpetc)?;
-    std::fs::write(tmpetc.join("someconfig.conf"), b"")?;
-    let tmproot = tmpetc.parent().unwrap();
-    let tmpvarlib = &tmproot.join("var/lib");
-    std::fs::create_dir_all(tmpvarlib)?;
-    std::fs::write(tmpvarlib.join("foo.log"), "foolog")?;
-    std::fs::write(tmpvarlib.join("bar.log"), "barlog")?;
-    std::fs::create_dir_all(tmproot.join("boot"))?;
-    let tmptar = fixture.path.join("testlayer.tar");
-    bash!(
-        "tar cf {tmptar} -C {tmproot} .",
-        tmptar = tmptar.as_str(),
-        tmproot = tmproot.as_str()
-    )?;
-    let src = tokio::fs::File::open(&tmptar).await?;
-    let r = ostree_ext::tar::write_tar(&fixture.destrepo, src, "layer", None).await?;
-    bash!(
-        "ostree --repo={repo} ls {layer_commit} /usr/etc/someconfig.conf >/dev/null",
-        repo = fixture.destrepo_path.as_str(),
+    fixture.dir.create_dir_all("tmproot/etc")?;
+    let tmproot = &fixture.dir.open_dir("tmproot")?;
+    let tmpetc = tmproot.open_dir("etc")?;
+    tmpetc.write("someconfig.conf", b"some config")?;
+    tmproot.create_dir_all("var/log")?;
+    let tmpvarlog = tmproot.open_dir("var/log")?;
+    tmpvarlog.write("foo.log", "foolog")?;
+    tmpvarlog.write("bar.log", "barlog")?;
+    tmproot.create_dir("boot")?;
+    let tmptar = "testlayer.tar";
+    bash_in!(fixture.dir, "tar cf ${tmptar} -C tmproot .", tmptar)?;
+    let src = fixture.dir.open(tmptar)?;
+    fixture.dir.remove_file(tmptar)?;
+    let src = tokio::fs::File::from_std(src.into_std());
+    let r = ostree_ext::tar::write_tar(fixture.destrepo(), src, "layer", None).await?;
+    bash_in!(
+        &fixture.dir,
+        "ostree --repo=dest/repo ls ${layer_commit} /usr/etc/someconfig.conf >/dev/null",
         layer_commit = r.commit.as_str()
     )?;
     assert_eq!(r.filtered.len(), 2);
@@ -407,6 +329,16 @@ async fn test_tar_write() -> Result<()> {
     Ok(())
 }
 
+#[tokio::test]
+async fn test_tar_write_tar_layer() -> Result<()> {
+    let fixture = Fixture::new()?;
+    let uncompressed_tar = tokio::io::BufReader::new(
+        async_compression::tokio::bufread::GzipDecoder::new(EXAMPLE_TAR_LAYER),
+    );
+    ostree_ext::tar::write_tar(&fixture.destrepo(), uncompressed_tar, "test", None).await?;
+    Ok(())
+}
+
 fn skopeo_inspect(imgref: &str) -> Result<String> {
     let out = Command::new("skopeo")
         .args(&["inspect", imgref])
@@ -415,12 +347,20 @@ fn skopeo_inspect(imgref: &str) -> Result<String> {
     Ok(String::from_utf8(out.stdout)?)
 }
 
+fn skopeo_inspect_config(imgref: &str) -> Result<oci_spec::image::ImageConfiguration> {
+    let out = Command::new("skopeo")
+        .args(&["inspect", "--config", imgref])
+        .stdout(std::process::Stdio::piped())
+        .output()?;
+    Ok(serde_json::from_slice(&out.stdout)?)
+}
+
 #[tokio::test]
 async fn test_container_import_export() -> Result<()> {
     let fixture = Fixture::new()?;
     let testrev = fixture
-        .srcrepo
-        .require_rev(TESTREF)
+        .srcrepo()
+        .require_rev(fixture.testref())
         .context("Failed to resolve ref")?;
 
     let srcoci_path = &fixture.path.join("oci");
@@ -435,13 +375,17 @@ async fn test_container_import_export() -> Result<()> {
                 .map(|(k, v)| (k.to_string(), v.to_string()))
                 .collect(),
         ),
-        cmd: Some(vec!["/bin/bash".to_string()]),
+        ..Default::default()
+    };
+    let opts = ostree_ext::container::ExportOpts {
+        copy_meta_keys: vec!["buildsys.checksum".to_string()],
+        ..Default::default()
     };
     let digest = ostree_ext::container::encapsulate(
-        &fixture.srcrepo,
-        TESTREF,
+        fixture.srcrepo(),
+        fixture.testref(),
         &config,
-        None,
+        Some(opts),
         &srcoci_imgref,
     )
     .await
@@ -452,6 +396,24 @@ async fn test_container_import_export() -> Result<()> {
     assert!(inspect.contains(r#""version": "42.0""#));
     assert!(inspect.contains(r#""foo": "bar""#));
     assert!(inspect.contains(r#""test": "value""#));
+    assert!(inspect.contains(
+        r#""buildsys.checksum": "41af286dc0b172ed2f1ca934fd2278de4a1192302ffa07087cea2682e7d372e3""#
+    ));
+    let cfg = skopeo_inspect_config(&srcoci_imgref.to_string())?;
+    // unwrap.  Unwrap.  UnWrap.  UNWRAP!!!!!!!
+    assert_eq!(
+        cfg.config()
+            .as_ref()
+            .unwrap()
+            .cmd()
+            .as_ref()
+            .unwrap()
+            .get(0)
+            .as_ref()
+            .unwrap()
+            .as_str(),
+        "/usr/bin/bash"
+    );
 
     let srcoci_unverified = OstreeImageReference {
         sigverify: SignatureSource::ContainerPolicyAllowInsecure,
@@ -466,7 +428,7 @@ async fn test_container_import_export() -> Result<()> {
         sigverify: SignatureSource::OstreeRemote("unknownremote".to_string()),
         imgref: srcoci_imgref.clone(),
     };
-    let r = ostree_ext::container::unencapsulate(&fixture.destrepo, &srcoci_unknownremote, None)
+    let r = ostree_ext::container::unencapsulate(fixture.destrepo(), &srcoci_unknownremote, None)
         .await
         .context("importing");
     assert_err_contains(r, r#"Remote "unknownremote" not found"#);
@@ -476,12 +438,11 @@ async fn test_container_import_export() -> Result<()> {
     opts.insert("gpg-verify", &true);
     opts.insert("custom-backend", &"ostree-rs-ext");
     fixture
-        .destrepo
+        .destrepo()
         .remote_add("myremote", None, Some(&opts.end()), gio::NONE_CANCELLABLE)?;
-    bash!(
-        "ostree --repo={repo} remote gpg-import --stdin myremote < {p}/gpghome/key1.asc",
-        repo = fixture.destrepo_path.as_str(),
-        p = fixture.srcdir.as_str()
+    bash_in!(
+        &fixture.dir,
+        "ostree --repo=dest/repo remote gpg-import --stdin myremote < src/gpghome/key1.asc",
     )?;
 
     // No remote matching
@@ -489,7 +450,7 @@ async fn test_container_import_export() -> Result<()> {
         sigverify: SignatureSource::OstreeRemote("myremote".to_string()),
         imgref: srcoci_imgref.clone(),
     };
-    let import = ostree_ext::container::unencapsulate(&fixture.destrepo, &srcoci_verified, None)
+    let import = ostree_ext::container::unencapsulate(fixture.destrepo(), &srcoci_verified, None)
         .await
         .context("importing")?;
     assert_eq!(import.ostree_commit, testrev.as_str());
@@ -499,7 +460,7 @@ async fn test_container_import_export() -> Result<()> {
     {
         let fixture = Fixture::new()?;
         let import =
-            ostree_ext::container::unencapsulate(&fixture.destrepo, &srcoci_unverified, None)
+            ostree_ext::container::unencapsulate(fixture.destrepo(), &srcoci_unverified, None)
                 .await
                 .context("importing")?;
         assert_eq!(import.ostree_commit, testrev.as_str());
@@ -531,8 +492,8 @@ async fn test_container_write_derive() -> Result<()> {
     let fixture = Fixture::new()?;
     let base_oci_path = &fixture.path.join("exampleos.oci");
     let _digest = ostree_ext::container::encapsulate(
-        &fixture.srcrepo,
-        TESTREF,
+        fixture.srcrepo(),
+        fixture.testref(),
         &Config {
             cmd: Some(vec!["/bin/bash".to_string()]),
             ..Default::default()
@@ -578,16 +539,16 @@ async fn test_container_write_derive() -> Result<()> {
         },
     };
     // There shouldn't be any container images stored yet.
-    let images = ostree_ext::container::store::list_images(&fixture.destrepo)?;
+    let images = ostree_ext::container::store::list_images(fixture.destrepo())?;
     assert!(images.is_empty());
 
     // Verify importing a derive dimage fails
-    let r = ostree_ext::container::unencapsulate(&fixture.destrepo, &derived_ref, None).await;
+    let r = ostree_ext::container::unencapsulate(fixture.destrepo(), &derived_ref, None).await;
     assert_err_contains(r, "Expected 1 layer, found 2");
 
     // Pull a derived image - two layers, new base plus one layer.
     let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
-        &fixture.destrepo,
+        fixture.destrepo(),
         &derived_ref,
         Default::default(),
     )
@@ -604,12 +565,12 @@ async fn test_container_write_derive() -> Result<()> {
     }
     let import = imp.import(prep).await?;
     // We should have exactly one image stored.
-    let images = ostree_ext::container::store::list_images(&fixture.destrepo)?;
+    let images = ostree_ext::container::store::list_images(fixture.destrepo())?;
     assert_eq!(images.len(), 1);
     assert_eq!(images[0], derived_ref.imgref.to_string());
 
     let imported_commit = &fixture
-        .destrepo
+        .destrepo()
         .load_commit(import.merge_commit.as_str())?
         .0;
     let digest = ostree_ext::container::store::manifest_digest_from_commit(imported_commit)?;
@@ -629,15 +590,15 @@ async fn test_container_write_derive() -> Result<()> {
     }
 
     // Parse the commit and verify we pulled the derived content.
-    bash!(
-        "ostree --repo={repo} ls {r} /usr/bin/newderivedfile >/dev/null",
-        repo = fixture.destrepo_path.as_str(),
+    bash_in!(
+        &fixture.dir,
+        "ostree --repo=dest/repo ls ${r} /usr/bin/newderivedfile >/dev/null",
         r = import.merge_commit.as_str()
     )?;
 
     // Import again, but there should be no changes.
     let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
-        &fixture.destrepo,
+        fixture.destrepo(),
         &derived_ref,
         Default::default(),
     )
@@ -654,7 +615,7 @@ async fn test_container_write_derive() -> Result<()> {
     std::fs::remove_dir_all(derived_path)?;
     std::fs::rename(derived2_path, derived_path)?;
     let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
-        &fixture.destrepo,
+        fixture.destrepo(),
         &derived_ref,
         Default::default(),
     )
@@ -674,26 +635,26 @@ async fn test_container_write_derive() -> Result<()> {
     // New commit.
     assert_ne!(import.merge_commit, already_present.merge_commit);
     // We should still have exactly one image stored.
-    let images = ostree_ext::container::store::list_images(&fixture.destrepo)?;
+    let images = ostree_ext::container::store::list_images(fixture.destrepo())?;
     assert_eq!(images[0], derived_ref.imgref.to_string());
     assert_eq!(images.len(), 1);
 
     // Verify we have the new file and *not* the old one
-    bash!(
+    bash_in!(
+        &fixture.dir,
         r#"set -x;
-         ostree --repo={repo} ls {r} /usr/bin/newderivedfile2 >/dev/null
-         test "$(ostree --repo={repo} cat {r} /usr/bin/newderivedfile)" = "newderivedfile v1"
-         if ostree --repo={repo} ls {r} /usr/bin/newderivedfile3 2>/dev/null; then
+         ostree --repo=dest/repo ls ${r} /usr/bin/newderivedfile2 >/dev/null
+         test "$(ostree --repo=dest/repo cat ${r} /usr/bin/newderivedfile)" = "newderivedfile v1"
+         if ostree --repo=dest/repo ls ${r} /usr/bin/newderivedfile3 2>/dev/null; then
            echo oops; exit 1
          fi
         "#,
-        repo = fixture.destrepo_path.as_str(),
         r = import.merge_commit.as_str()
     )?;
 
     // And there should be no changes on upgrade again.
     let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
-        &fixture.destrepo,
+        fixture.destrepo(),
         &derived_ref,
         Default::default(),
     )
@@ -714,7 +675,7 @@ async fn test_container_write_derive() -> Result<()> {
         None,
         gio::NONE_CANCELLABLE,
     )?;
-    ostree_ext::container::store::copy(&fixture.destrepo, &destrepo2, &derived_ref).await?;
+    ostree_ext::container::store::copy(fixture.destrepo(), &destrepo2, &derived_ref).await?;
 
     let images = ostree_ext::container::store::list_images(&destrepo2)?;
     assert_eq!(images.len(), 1);
@@ -734,9 +695,10 @@ async fn test_container_write_derive() -> Result<()> {
 async fn test_container_import_export_registry() -> Result<()> {
     let tr = &*TEST_REGISTRY;
     let fixture = Fixture::new()?;
+    let testref = fixture.testref();
     let testrev = fixture
-        .srcrepo
-        .require_rev(TESTREF)
+        .srcrepo()
+        .require_rev(testref)
         .context("Failed to resolve ref")?;
     let src_imgref = ImageReference {
         transport: Transport::Registry,
@@ -747,7 +709,7 @@ async fn test_container_import_export_registry() -> Result<()> {
         ..Default::default()
     };
     let digest =
-        ostree_ext::container::encapsulate(&fixture.srcrepo, TESTREF, &config, None, &src_imgref)
+        ostree_ext::container::encapsulate(fixture.srcrepo(), testref, &config, None, &src_imgref)
             .await
             .context("exporting to registry")?;
     let mut digested_imgref = src_imgref.clone();
@@ -757,7 +719,7 @@ async fn test_container_import_export_registry() -> Result<()> {
         sigverify: SignatureSource::ContainerPolicyAllowInsecure,
         imgref: digested_imgref,
     };
-    let import = ostree_ext::container::unencapsulate(&fixture.destrepo, &import_ref, None)
+    let import = ostree_ext::container::unencapsulate(fixture.destrepo(), &import_ref, None)
         .await
         .context("importing")?;
     assert_eq!(import.ostree_commit, testrev.as_str());
@@ -766,15 +728,12 @@ async fn test_container_import_export_registry() -> Result<()> {
 
 #[test]
 fn test_diff() -> Result<()> {
-    let cancellable = gio::NONE_CANCELLABLE;
-    let tempdir = tempfile::tempdir()?;
-    let tempdir = Utf8Path::from_path(tempdir.path()).unwrap();
-    let repopath = &generate_test_repo(tempdir)?;
-    update_repo(repopath)?;
-    let from = &format!("{}^", TESTREF);
-    let repo = &ostree::Repo::open_at(libc::AT_FDCWD, repopath.as_str(), cancellable)?;
+    let mut fixture = Fixture::new()?;
+    fixture.update()?;
+    let from = &format!("{}^", fixture.testref());
+    let repo = fixture.srcrepo();
     let subdir: Option<&str> = None;
-    let diff = ostree_ext::diff::diff(repo, from, TESTREF, subdir)?;
+    let diff = ostree_ext::diff::diff(repo, from, fixture.testref(), subdir)?;
     assert!(diff.subdir.is_none());
     assert_eq!(diff.added_dirs.len(), 1);
     assert_eq!(diff.added_dirs.iter().next().unwrap(), "/usr/share");
@@ -782,7 +741,7 @@ fn test_diff() -> Result<()> {
     assert_eq!(diff.added_files.iter().next().unwrap(), "/usr/bin/newbin");
     assert_eq!(diff.removed_files.len(), 1);
     assert_eq!(diff.removed_files.iter().next().unwrap(), "/usr/bin/foo");
-    let diff = ostree_ext::diff::diff(repo, from, TESTREF, Some("/usr"))?;
+    let diff = ostree_ext::diff::diff(repo, from, fixture.testref(), Some("/usr"))?;
     assert_eq!(diff.subdir.as_ref().unwrap(), "/usr");
     assert_eq!(diff.added_dirs.len(), 1);
     assert_eq!(diff.added_dirs.iter().next().unwrap(), "/share");
